/**
  *  \file client/screens/unpackscreen.cc
  *  \brief Unpack screen
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens/unpackscreen.h"
#include "client/dialogs/consolelog.h"
#include "client/widgets/setselect.h"
#include "cpluslib/format.h"
#include "game/gamedirs.h"
#include "game/rst.h"
#include "game/storage/overview.h"
#include "game/struct.h"
#include "game/unpack.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/widgets/multilinestatic.h"
#include "util/console.h"
#include "util/except.h"
#include "util/nls.h"
#include "client/widgets/imagepanel.h"
#include "ui/widgets/alignwidget.h"
#include "resmgr/resmgr.h"
#include "ui/widgets/button.h"
#include "client/widgets/helpwidget.h"
#include "ui/richdocument.h"
#include "ui/richhelp.h"
#include "util/progress.h"
#include "game/backup.h"
#include "util/conf.h"
#include "game/pref.h"
#include "ui/commands.h"
#include "game/trn.h"
#include "game/trnapply.h"
#include "ui/widgets/quit.h"

namespace {
    enum { WIDTH = 600 };

    class WUnpackWindow : public WImagePanel {
        ObjectHolder h;
        WPlayerSetSelector& set_sel;
        GDirectoryOverview& ov;

        void init();
     public:
        WUnpackWindow(WPlayerSetSelector& set_sel, GDirectoryOverview& ov);
        ~WUnpackWindow();

        void onOK();
    };

    class AskingAnswerProvider : public AnswerProvider {
     public:
        bool ask(int, string_t q)
            {
                return UIMessageBox(q, _("Unpack")).doYesNoDialog();
            }
    };
}

/* Configuration */
static ValueEnumParser parse_unpack_format("DOS,Windows");
static ConfigIntOption opt_unpack_format   (getUserPreferences(), "Unpack.Format", parse_unpack_format, 1);
static ConfigIntOption opt_unpack_fixerrors(getUserPreferences(), "Unpack.FixErrors", ValueBoolParser::instance, 1);
static ConfigIntOption opt_unpack_targetext(getUserPreferences(), "Unpack.TargetExt", ValueBoolParser::instance, 0);


WUnpackWindow::WUnpackWindow(WPlayerSetSelector& set_sel, GDirectoryOverview& ov)
    : WImagePanel(_("Unpack"), requestPixmap(ResId("bg.unpack", ui_root->getExtent().w))),
      set_sel(set_sel),
      ov(ov)
{
    init();
}

WUnpackWindow::~WUnpackWindow()
{ }

void
WUnpackWindow::init()
{
    UIGroup& main = h.add(new UIGroup(UIVBoxLayout::instance5));

    UIRichDocument& doc = h.add(new UIRichDocument(GfxPoint(WIDTH, 1), 0));
    doc.getDocument().setPageWidth(WIDTH);
    doc.getDocument().add(parseRichText(_("The <em>Unpack</em> function will unpack result files into "
                                          "individual files. After unpacking, you can play your "
                                          "turn using all VGAP utilities of your choice. When "
                                          "done with the turn, use the Maketurn function to "
                                          "generate TRN files to send to the host.<br/>"
                                          "Select the files you want to unpack using <kbd>Space</kbd> "
                                          "or using the mouse.<br/>"
                                          "Note that PCC2 doesn't require you to unpack a result "
                                          "file prior to playing, but most other utilities do.")));
    doc.getDocument().finish();
    doc.adjustToDocumentSize();

    main.add(h.add(new UIMultilineStatic(FONT_NORMAL, _("Choose files to unpack:"), WIDTH)));
    main.add(set_sel);
    main.add(doc);

    UIButton& btn_ok     = h.add(new UIButton(1, _("OK"), SDLK_RETURN));
    UIButton& btn_cancel = h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE));

    btn_ok.sig_fired.add(this, &WUnpackWindow::onOK);
    btn_cancel.addStop();

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(btn_ok);
    g.add(btn_cancel);
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(_("Help"), 'h')));
    main.add(g);

    add(h.add(new UIAlignedWidget(main, GfxPoint(1, 1), GfxPoint(10, 10))));
    add(h.add(new WHelpWidget("pcc2:unpack")));
    add(h.add(new UIQuit(0)));
    setExtent(ui_root->getExtent());
    setState(st_Modal, true);
}

void
WUnpackWindow::onOK()
{
    // Fetch selection
    GPlayerSet selected = set_sel.getSelectedPlayers();
    if (selected.empty())
        return;

    // Show console. First, determine size. Default is 7 lines per player.
    int nlines = 0;
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        if (selected.contains(i) && nlines < 14)
            nlines += 7;
    WConsoleLogWindow log(_("Unpack"), nlines);
    log.scrollAfterEnd();
    log.open();

    // Unpack selection
    AskingAnswerProvider aap;
    ProgressMonitor pm;
    GUnpacker theUnpacker;
    theUnpacker.setFixErrors(opt_unpack_fixerrors());
    theUnpacker.setCreateTargetExt(opt_unpack_targetext());
    theUnpacker.setFormat(opt_unpack_format() ? GUnpacker::WindowsFormat : GUnpacker::DosFormat);
    theUnpacker.setAnswerProvider(aap);
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (selected.contains(i)) {
            // Loda turn file, if any
            std::auto_ptr<GUnpackModifier> modif;
            if (ov.getPlayerFlags(i) & ov.have_TRN) {
                UIMessageBox box((format(_("A turn file has been found for player %d. "
                                           "Should this turn file be used to restore the state before the last Maketurn?"),
                                         i)),
                                 _("Unpack"));
                UIKeyString y("Yes"), n("No");
                box.addButton(cm_Enter, y.getString(), y.getKey());
                box.addButton(1,        n.getString(), n.getKey());
                box.addButton(cm_Escape, _("Cancel"), SDLK_ESCAPE);
                box.addKey(cm_Enter, ' ');
                int response = box.run();
                if (response == cm_Escape) {
                    // Cancel
                    break;
                }
                if (response == cm_Enter) {
                    // Yes
                    string_t trnName = format("player%d.trn", i);
                    try {
                        Ptr<Stream> str = game_file_dir->openFileNT(trnName, Stream::C_READ);
                        Ptr<GTurnfile> trn = new GTurnfile(*str);
                        modif.reset(new GTurnfileApply(trn, opt_unpack_format()));
                    }
                    catch (FileProblemException& e) {
                        console << e.getFileName() << ": " << e.what() << "\n";
                    }
                    catch (std::exception& e) {
                        console << trnName << ": " << e.what() << "\n";
                    }
                }
            }

            string_t filename = format("player%d.rst", i);
            try {
                Ptr<Stream> rst = game_file_dir->openFileNT(filename, Stream::C_READ);
                GResultFile rstfile(*rst);
                console.write(LOG_INFO, format(_("=== Unpacking player %d... ==="), i));
                theUnpacker.unpack(rstfile, i, pm, modif.get());

                // Backup
                GBackupFileTemplate tpl;
                tpl.setGameDirectory(game_dir_name);
                tpl.setPlayerNumber(i);
                tpl.setTurnNumber(theUnpacker.getTurnNumber());
                rst->seek(0);
                tpl.copyFile(opt_BackupResult(), *rst, pm);
            }
            catch (FileProblemException& e) {
                console << e.getFileName() << ": " << e.what() << "\n";
            }
            catch (std::exception& e) {
                console << filename << ": " << e.what() << "\n";
            }
        }
    }
    log.waitForConfirmation();
    stop(1);
}    

/** Do unpack screen. Asks the user for a set of players in the
    game_file_dir, and unpacks their results.
    \param defaultSet players selected by default */
void
doUnpack(GPlayerSet defaultSet)
{
    // VBox
    //   UIMultilineStatic
    //   WPlayerSetSelector
    //   HBox
    //     UIButton "OK"
    //     UIButton "Cancel"

    /* Check what's available */
    GDirectoryOverview ov;
    ov.scan(*game_file_dir, true);
    GPlayerSet avail = ov.getPlayersWhere(ov.have_RST);
    if (avail.empty()) {
        messageBox(_("This directory contains no files to unpack."),
                   _("Unpack"));
        return;
    }

    /* Build player selector */
    WPlayerSetSelector psel(1000, NUM_PLAYERS, WIDTH);
    psel.setSelectablePlayers(avail);
    psel.setSelectedPlayers(avail & defaultSet);
    psel.scrollTo(0, 1);
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (ov.getPlayerFlags(i) & ov.have_RST) {
            if (ov.getPlayerFlags(i) & ov.have_TRN) {
                psel.setInfo(i, _("RST + TRN"));
            } else {
                psel.setInfo(i, _("RST"));
            }
        }
    }

    /* Build window */
    WUnpackWindow win(psel, ov);
    ui_root->centerWidget(win);
    ui_root->runModal(win);
}
