/**
  *  \file client/tiles/csheader.cc
  *  \brief WControlScreenHeaderTile tile
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/csheader.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "client/wcolorframe.h"
#include "ui/widgets/imagewidget.h"
#include "ui/widgets/button.h"
#include "int/process.h"
#include "int/notify.h"


/** Get frame color for Control Screen ("Auto" button).
    - green for presence of auto task
    - red if auto task has an error (pending notify)
    - yellow for presence of another process */
static TFrameColor
getControlScreenFrameColor(GObject* obj, IntExecutionContext::ProcessKind pk)
{
    const IntProcessList& pl = getProcessList();
    bool any = false;
    for (IntProcessList::iterator i = pl.begin(); i != pl.end(); ++i) {
        // Check for a process which is started from this object, and
        // which is currently runnable/suspended/frozen. Those are the
        // states usually assumed by auto tasks or long-running scripts.
        // Running scripts do not count here, as they are usually (but
        // not always!) temporary UI processes.
        if (((*i)->getState() == IntExecutionContext::psRunnable
             || (*i)->getState() == IntExecutionContext::psSuspended
             || (*i)->getState() == IntExecutionContext::psFrozen)
            && (*i)->getInvokingObject() == obj)
        {
            any = true;
            if ((*i)->getProcessKind() == pk) {
                // Found the auto task
                IntNotificationMessage* nm = (*i)->getNotificationMessage();
                if (nm != 0 && !nm->isConfirmed()) {
                    return cf_Red;
                } else {
                    return cf_Green;
                }
            }
        }
    }
    if (any) {
        return cf_Yellow;
    } else {
        return cf_None;
    }
}

/** Get frame color for Auto Task Screen ("CScr" button).
    Only show presence of errors here. */
static TFrameColor
getAutoTaskFrameColor(GObject* obj, IntExecutionContext::ProcessKind pk)
{
    IntExecutionContext* proc = getProcessByObject(obj, pk);
    if (proc != 0) {
        IntNotificationMessage* nm = proc->getNotificationMessage();
        if (nm != 0 && !nm->isConfirmed()) {
            return cf_Red;
        } else {
            return cf_None;
        }
    } else {
        return cf_None;
    }
}

/** Constructor.
    \param selection object selection to observe
    \param buttons the buttons you want to be displayed
    \param pk Process kind for auto tasks */
WControlScreenHeaderTile::WControlScreenHeaderTile(GObjectSelection& selection, int buttons, IntExecutionContext::ProcessKind pk)
      // FIXME: magic numbers
    : WObjectSelectionObserverTile(GfxPoint(315, 145), selection), buttons(buttons),
      pk(pk)
{
    init();
}

void
WControlScreenHeaderTile::init()
{
    setLayout(LeftJustified, 5, false);
    addControl(GfxRect( 5-5,  45, 30, 25),
               holder.add(new UIButton("F1", SDLK_F1)));
    addControl(GfxRect(40-5,  45, 30, 25),
               holder.add(new UIButton("F2", SDLK_F2)));
    addControl(GfxRect( 5-5,  75, 30, 25),
               holder.add(new UIButton("F3", SDLK_F3)));
    addControl(GfxRect(40-5,  75, 30, 25),
               holder.add(new UIButton("F4", SDLK_F4)));
    addControl(GfxRect( 5-5, 105, 30, 25),
               holder.add(new UIButton("F6", SDLK_F6)));
    addControl(GfxRect(40-5, 105, 30, 25),
               holder.add(new UIButton("F7", SDLK_F7)));
    addControl(GfxRect(265-5, 114, 45, 25),
               holder.add(new UIButton("ESC", SDLK_ESCAPE)));
    addControl(GfxRect(230-5, 114, 25, 25),
               holder.add(new UIButton("H", 'h')));
    addControl(GfxRect(285-5, 45, 25, 25),
               holder.add(new UIButton(cm_CS_Xchg, "X", 'x')).addKey());
    addControl(GfxRect(75-5, 45, 20, 25),
               holder.add(new UIButton("\xE2\x86\x91", '-')));
    addControl(GfxRect(75-5, 75, 20, 25),
               holder.add(new UIButton("\xE2\x86\x93", '+')));
//     if (buttons & bSend)
//         addControl(GfxRect(75-5, 105, 20, 25),
//                    holder.add(new UIButton("I", 'i')));

    if (buttons & (bAuto | bCScr)) {
        auto_cf = &holder.add(new WColorFrame());
        if (buttons & bAuto) {
            auto_cf->add(holder.add(new UIButton("Auto", SDLK_RETURN)));
        } else {
            auto_cf->add(holder.add(new UIButton("CScr", SDLK_RETURN)));
        }
        addControl(GfxRect(228-5, 43, 54, 29), *auto_cf);
    } else {
        auto_cf = 0;
    }

    if (buttons & bCycle) {
        addControl(GfxRect(275-5, 75, 35, 25),
                   holder.add(new UIButton("\xEE\x85\x80", SDLK_TAB)));
    } else if (buttons & bJoin) {
        addControl(GfxRect(285-5, 75, 25, 25), holder.add(new UIButton("J", 'j')));
    }
    if (buttons & bAdd)
        addControl(GfxRect(230-5, 75, 40, 25),
                   holder.add(new UIButton("Add", SDLK_INSERT)));
    if (buttons & bName)
        addControl(GfxRect(315-20-5, 0, 20, 20),
                   holder.add(new UIButton("N", 'n')));

    // Image
    image_cf = &holder.add(new WColorFrame());
    image = &holder.add(new UIImageWidget(105, 93));
    addControl(GfxRect(108-5, 45, 107, 95), *image_cf);
    image_cf->add(*image);
}

void
WControlScreenHeaderTile::drawData(GfxCanvas& can)
{
    // FIXME: magic numbers
    int wi = (buttons & bName) ? extent.w-20 : extent.w;

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.setSolidBackground();

    ctx.setColor(UIColor::tc_Heading);
    ctx.useFont(*fonts[FONT_TITLE]);
    outTextF(ctx, 0, 0, wi, getTitle());

    ctx.setColor(UIColor::tc_Yellow);
    ctx.useFont(*fonts[FONT_NORMAL]);
    outTextF(ctx, 0, 25, extent.w, getSubtitle());

    image->setImage(ResManager::getInstance().getPixmap(getPictureId()));
    image_cf->setColor(getCurrentObject()->isMarked() ? cf_Yellow : cf_None);

    if (auto_cf) {
        TFrameColor fc = cf_None;
        if (buttons & bAuto) {
            fc = getControlScreenFrameColor(getCurrentObject(), pk);
        }
        if (buttons & bCScr) {
            fc = getAutoTaskFrameColor(getCurrentObject(), pk);
        }
        auto_cf->setColor(fc);
    }
}
