/**
  *  \file client/widgets/vcrhulllist.cc
  *  \brief WVcrHullList widget
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/vcrhulllist.h"
#include "util/nls.h"
#include "ui/util.h"
#include "game/spec.h"
#include "game/vcrdb.h"
#include "game/vcrobj.h"

/** Constructor.
    \param db VCR database to observe
    \param slot VCR slot to display initially
    \param side VCR side to display initially */
WVcrHullList::WVcrHullList()
    : UIListbox(0, 3, 20*font_ems[FONT_NORMAL], 0),
      pEntry(0),
      side(0)
{
    buildList();
}

WVcrHullList::~WVcrHullList()
{ }

void
WVcrHullList::drawPart(GfxCanvas& can, int from, int to)
{
    /* The placeholder code used to be in drawContent, but UIListbox occasionally calls
       drawPart when a part of the list must be redrawn, so we have no alternative to
       doing it here. */
    if (pEntry == 0) {
        /* This is not a regular mode of operation, so there's no need
           to invent a fancy text. */
        drawPlaceholder(can, string_t());
    } else if (pEntry->getObject(side, false).isPlanet()) {
        drawPlaceholder(can, _("This is a planet."));
    } else if (possible_types.empty()) {
        drawPlaceholder(can, _("The type of this ship cannot be determined."));
    } else {
        GfxContext ctx;
        ctx.useCanvas(can);
        ctx.useFont(*fonts[FONT_NORMAL]);
        for (int i = from; i < to; ++i) {
            int slot = i + getTop();
            GfxRect area;
            prepareEntry(ctx, ColorStyle, i, &area);
            if (slot < int(possible_types.size())) {
                outTextF(ctx, area.x+5, area.y, area.w-5, getHull(possible_types[i]).getName());
            }
        }
    }
}

/** Get currently-selected hull number.
    \return hull number, 0 if none matches or this is a planet */
int
WVcrHullList::getCurrentHull() const
{
    if (pEntry == 0 || pEntry->getObject(side, false).isPlanet() || getCurrent() >= int(possible_types.size())) {
        return 0;
    } else {
        return possible_types[getCurrent()];
    }
}

/** Get picture for currently-selected item.
    \return picture of hull if any selected, picture from VCR otherwise, 0 if this is a planet */
int
WVcrHullList::getCurrentPicture() const
{
    if (pEntry == 0 || pEntry->getObject(side, false).isPlanet()) {
        return 0;
    } else if (getCurrent() >= int(possible_types.size())) {
        return pEntry->getObject(side, false).getPicture();
    } else {
        return getHull(possible_types[getCurrent()]).getInternalPictureNumber();
    }
}

/** Set side to display. This will raise sig_moved when anything changes.
    \param side 0 or 1 */
void
WVcrHullList::setSide(int side)
{
    if (side != this->side) {
        this->side = side;
        buildList();
    }
}

/** Get side to display. \see setSide */
int
WVcrHullList::getSide() const
{
    return side;
}

/** Set VCR to display. */
void
WVcrHullList::setVcrEntry(GVcrEntry& entry, int side)
{
    if (&entry != this->pEntry || side != this->side) {
        this->pEntry = &entry;
        this->side = side;
        buildList();
    }
}

/** Get current VCR entry. */
GVcrEntry&
WVcrHullList::getVcrEntry()
{
    return *pEntry;
}

/** Draw placeholder text. This is used when the list is empty.
    \param text Text to display. Will be word-wrapped, and can contain "\n". */
void
WVcrHullList::drawPlaceholder(GfxCanvas& can, string_t text)
{
    int width = getExtent().w * 15/16;
    int nlines = countLines(*fonts[FONT_NORMAL], width, text);

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.useColorScheme(getSkin());
    ctx.setTextAlign(1, 0);

    drawBackground(ctx, getExtent());
    outTextFormatted(ctx, getExtent().x + getExtent().w/2, getExtent().y + getExtent().h/2 - nlines*font_heights[FONT_NORMAL]/2,
                     width, text);
}

/** Build list. This will reset the list and raise sig_moved. */
void
WVcrHullList::buildList()
{
    possible_types.clear();

    if (pEntry != 0) {
        for (int i = 1, e = getNumHulls(); i < e; ++i) {
            if (pEntry->getObject(side, false).canBeHull(i)) {
                possible_types.push_back(i);
            }
        }
    }

    /* Reset and redraw everything. */
    setCurrent(0);
    setTop(0);
    sig_moved.raise(getId());
    drawWidget();
}
