/**
  *  \file game/actions/basebuild.h
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_ACTIONS_BASEBUILD_H
#define PCC_V2_GAME_ACTIONS_BASEBUILD_H

#include "game/planet.h"
#include "cpluslib/sigslot.h"
#include "game/actions/cargocost.h"
#include "game/cost.h"

/** Abstract starbase build action executor.
    Derived classes receive requests from starbase build actions to
    produce computations or perform the actual build. */
class GStarbaseBuildExecuter : public Object {
 public:
    /** Change tech level.
        \param area which tech level to change
        \param value new value */
    virtual void setTech(TTechLevel area, int value) = 0;

    /** Change hull storage.
        \param index hull slot [1,NUM_HULLS_PER_PLAYER]
        \param value new number of hulls
        \param free additional number of (previously present) hulls consumed by action at no cost */
    virtual void setHullSlot(int index, int value, int free) = 0;

    /** Change engine storage.
        \param index engine slot [1,NUM_ENGINES]
        \param value new number of engines
        \param free additional number of (previously present) engines consumed by action at no cost */
    virtual void setEngine(int index, int value, int free) = 0;

    /** Change beam storage.
        \param index beam slot [1,NUM_BEAMS]
        \param value new number of beams
        \param free additional number of (previously present) beams consumed by action at no cost */
    virtual void setBeam(int index, int value, int free) = 0;

    /** Change torpedo launcher storage.
        \param index torpedo slot [1,NUM_TORPS]
        \param value new number of torpedo launchers
        \param free additional number of (previously present) launchers consumed by action at no cost */
    virtual void setLauncher(int index, int value, int free) = 0;

    /** Account for a foreign hull. Intended for price computations involving hulls the player
        can not build. Such a change cannot be committed.
        \param index hull slot [1,NUM_HULLS]
        \param count number of hulls to build
        \param free additional number of (previously present) hulls consumed by action at no cost */
    virtual void accountHull(int number, int count, int free) = 0;
};

/** Abstract starbase build action.
    This is not a transaction to make it possible to manipulate it without
    having a transaction around all the time which blocks resources on the planet. */
class GStarbaseBuildAction : public Object {
    GPlanet& planet;
 public:
    GStarbaseBuildAction(GPlanet& planet);
    GPlanet& getPlanet();
    const GPlanet& getPlanet() const;
    GCost getCost();

    // Derived class:
    virtual void perform(GStarbaseBuildExecuter& exec) = 0;
    virtual GError getError() = 0;
    Signal0 sig_changed;
};

/** Starbase build transaction. This turns a GStarbaseBuildAction into
    a GCargoCostTransaction. */
class GStarbaseBuildTransaction : public GCargoCostTransaction {
    GStarbaseBuildAction& action;
 public:
    GStarbaseBuildTransaction(GStarbaseBuildAction& action, GCargoContainer& container);
    ~GStarbaseBuildTransaction();
    void update();
    GError getError();
    Signal0 sig_cost_changed;

 protected:
    // GCargoCostTransaction:
    void commit();
    void cancel();

    SignalConnection conn;
    SignalConnection conn2;
};



inline
GStarbaseBuildAction::GStarbaseBuildAction(GPlanet& planet)
    : planet(planet)
{ }

inline GPlanet&
GStarbaseBuildAction::getPlanet()
{
    return planet;
}

inline const GPlanet&
GStarbaseBuildAction::getPlanet() const
{
    return planet;
}

#endif
