/**
  *  \file game/cargo.cc
  *  \brief Cargo Reservation/Arbitration Primitives
  *
  *  PCC2 supports "concurrent" access to cargo containers. Hence,
  *  we use a transaction-like scheme to ensure correct behaviour. 
  *  Each object has a cargo arbiter (GCargoArbiter), and a number of cargo
  *  traits (GCargoContainer). Cargo which is removed from the object
  *  is first reserved, and not added to the target object until the
  *  transaction is committed. We don't support "real", preemptive
  *  concurrency, though.
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/cargo.h"
#include "cpluslib/assert.h"
#include "cpluslib/misc.h"
#include "game/spec.h"

/***************************** GCargoArbiter *****************************/

struct GCargoArbiter::Impl {
    /** Reference count. Equals the number of incUseCount() calls
        which have not been taken back using decUseCount() yet. */
    int ref_count;
    /** Reserved cargo which is not available. */
    int32 reserved_cargo[el_COUNT];

    Impl();
};

/** Constructor. */
GCargoArbiter::Impl::Impl()
    : ref_count(0)
{
    std::fill_n(reserved_cargo, int(el_COUNT), 0);
}

GCargoArbiter::~GCargoArbiter() throw()
{
    delete data;
}

/** Increment use count. Every piece of code which possibly does a
    cargo reservation must call this. */
void
GCargoArbiter::incUseCount()
{
    if (!data)
        data = new Impl();
    ++data->ref_count;
}

/** Decrement use count. You may no longer call changeReservedCargo()
    after calling decUseCount(). */
void
GCargoArbiter::decUseCount() throw()
{
    if (!--data->ref_count) {
        delete data;
        data = 0;
    }
}

/** Check whether cargo arbiter is in use. */
bool
GCargoArbiter::isInUse() const
{
    return data != 0;
}

/** Get reserved cargo. User interface should not include this amount
    in display. Every transaction must assume that this amount is not
    available (except for cargo known to reserved by the transaction
    itself). */
int32
GCargoArbiter::getReservedCargo(GCargoType t) const throw()
{
    return data ? data->reserved_cargo[t] : 0;
}

/** Change cargo reservation.
    \param t cargo type
    \param amount >0 to reserve cargo, <0 to give back
    \returns new amount of reserved cargo. */
int32
GCargoArbiter::changeReservedCargo(GCargoType t, int32_t amount) throw()
{
    if (data) {
        ASSERT(data->reserved_cargo[t] + amount >= 0);
        data->reserved_cargo[t] += amount;
        return data->reserved_cargo[t];
    } else {
        ASSERT(amount == 0);
        return 0;
    }
}

/**************************** GCargoContainer ****************************/

/*! \class GCargoContainer

    This is the base class for everything that deals with cargo
    (minerals, sup, mc, fighters, torps, clans). It is used for cargo
    transfer and some other things.

    Derived classes provide access to the underlying object.
    - getCargoOnObject() yields the current amount on the object, where
      reserved stuff has already been subtracted (that is, the
      amount which can still be safely subtracted).
    - changeCargoOnObject() changes the current physical amount.
    - changeReservedCargo() changes the current reserved amount.
    - getMaxCargo() yields the maximum amount which can appear on the
      object.

    The public manipulation interface uses these functions to provide
    safe cargo amount manipulation. Using commit(), you can commit the
    changes; cancel() takes them back.

    \todo FIXME: a problem which can not be solved is to test whether a
    package of multiple cargo types can fit onto a container without
    modifying it, e.g. to test whether I can add "50TDM 2800$". Things
    I want to do with this:
    - "load minerals for as many torps as possible" (mkt support)
    - clearing the hold space in multi-cargo transfer.

    It is still possibly doing this item-by-item, but that is
    cumbersome.

    Interfaces that come to mind (public virtual functions):
    - bool canAdd(GCargoContainer& other): return true iff other
      can be added to *this. This solves problem 2; problem 1 can
      be solved using binary intersection tests.
    - int divide(GCargoContainer& other): how many times other
      can be added to *this. Elegant at first, this one has problems
      handling "infinite" containers (planets). */

/** Constructor. Creates a cargo container which did not yet
    add/remove anything from the underlying object. */
GCargoContainer::GCargoContainer()
    : flags(0)
{
    std::fill_n(diffs, countof(diffs), 0);
}

GCargoContainer::~GCargoContainer()
{ }

/** Return current amount of cargo. This is the value to be presented
    to the user. It includes others' removals but not their additions,
    and it includes our removals and additions. Therefore it is safe
    in the sense that the value returned from this function can be
    removed from the object without problems. */
int32
GCargoContainer::getCargo(GCargoType type)
{
    if (diffs[type] < 0) {
        /* we have removed some cargo from this object. Hence, getAmount()
           will include our reservation. */
        return getCargoOnObject(type);
    } else {
        /* we are receiving cargo. getAmount() will include others' reservations. */
        return getCargoOnObject(type) + diffs[type];
    }
}

/** Change cargo.
    \param type Type of cargo to change
    \param amount Amount to change. Positive to add to this object, negative to remove
    \param partial true to accept partial move, false to accept only complete move
    \return amount moved. If partial=true, this can be any value equal to or smaller in
    magnitude than the amount. If partial=false, this will be either 0 or amount. */
int32_t
GCargoContainer::changeCargo(const GCargoType type, int32_t amount, const bool partial)
{
    if (amount < 0) {
        // Remove
        ASSERT(!isFinished());
        const int32_t have = getCargo(type);
        if (-amount > have) {
            // We're trying to remove more than we have. Exit without any
            // change if we're not permitted to do partial change, or if
            // we don't have anything.
            if (!partial || have == 0)
                return 0;
            amount = -have;
        }

        // Cargo removal is split into giving back previously-received cargo, and reservation of new cargo.
        const int32_t previouslyReceived = std::max(std::min(diffs[type], -amount), int32_t(0));
        const int32_t newlyReserved = -amount - previouslyReceived;

        // Now reserve the remainder
        changeReservedCargo(type, newlyReserved);

        diffs[type] += amount;
        sig_changed.raise();
        return amount;
    } else if (amount > 0) {
        // Add. First check capacity.
        ASSERT(!isFinished());
        int32_t room = getMaxCargo(type) - getCargo(type);

        // Check overload
        if (room <= 0) {
            return 0;
        }

        // Partial move
        if (amount > room) {
            if (!partial)
                return 0;
            amount = room;
        }

        // Undo reservation, if any
        const int32_t previouslyReserved = std::max(std::min(-diffs[type], amount), int32_t(0));
        changeReservedCargo(type, -previouslyReserved);

        // Update diffs
        diffs[type] += amount;
        sig_changed.raise();
        return amount;
    } else {
        // No change
        return 0;
    }
}

/** Check for change.
    \retval true if this container lost or received some cargo
    \retval false if this container was not modified by the current transaction */
bool
GCargoContainer::isChanged() const
{
    for (GCargoType i = el_FIRST; i < el_COUNT; ++i) {
        if (diffs[i] != 0) {
            return true;
        }
    }
    return false;
}

/** Cancel transaction. This will undo our reservations. */
void
GCargoContainer::cancel()
{
    /* cancelling a transaction means canceling our reservations */
    for (GCargoType i = el_FIRST; i < el_COUNT; ++i) {
        if (diffs[i] < 0) {
            changeReservedCargo(i, diffs[i]);
            diffs[i] = 0;
        }
    }
}

/** Commit transaction. This will perform all changes. */
void
GCargoContainer::commit()
{
    /* commit means to cancel our reservations *and* apply our changes */
    for (GCargoType i = el_FIRST; i < el_COUNT; ++i) {
        if (diffs[i] < 0)
            changeReservedCargo(i, diffs[i]);
        if (diffs[i] != 0)
            changeCargoOnObject(i, diffs[i]);
        diffs[i] = 0;
    }
}
