/**
  *  \file game/coord.cc
  *  \brief Game Coordinate System Handling
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <ostream>
#include <cmath>
#include "game/coord.h"
#include "game/config.h"
#include "game/host.h"
#include "util/misc.h"
#include "util/modinit.h"
#include "util/conf.h"
#include "game/pref.h"

/** Global starchart configuration. */
GChartConfiguration chart_conf;

/** Circular Wrap Precision option. Determines how many iterations are performed to
    find the outside location of an object. */
static ConfigIntOption opt_circular_precision(getUserPreferences(), "Chart.Circle.Precision", ValueIntParser::instance, 2);

/** Circular Wrap Excess option. Determines the maximum size of the outside area on
    the circular wrapped map. This is limited to avoid math problems due to the heavy
    distortion. */
static ConfigIntOption opt_circular_excess(getUserPreferences(), "Chart.Circle.Outside", ValueIntParser::instance, 501);

/** Wrap mode. */
static ValueEnumParser parse_map_kind("flat,wrapped,circular");
static ConfigIntOption opt_map_kind(getUserPreferences(), "Chart.Geo.Mode", parse_map_kind, 0);

/** Map center. */
static ConfigIntArrayOption<2> opt_map_center(getUserPreferences(), "Chart.Geo.Center", ValueIntParser::instance, 2000);

/** Map size. */
static ConfigIntArrayOption<2> opt_map_size(getUserPreferences(), "Chart.Geo.Size", ValueIntParser::instance, 2000);


/* Map Images:
     1 2 3
     4 0 5
     6 7 8 */
static const int8_t image_dx[] = {  0, -1,  0, +1, -1, +1, -1,  0, +1 };
static const int8_t image_dy[] = {  0, -1, -1, -1,  0,  0, +1, +1, +1 };



/** Get exact mapping for a point with circular wrap.

    Circular wrap cannot be inverted with a simple formula:

    Let x be the position where our ship is, x' be the position where PWrap
    maps it to, assuming that x is outside the wrap circle. Here, we know x'
    and want to know which x we need.

    We have
      x' = Trunc(CenterX + nr*sin(angle) + 0.5)
    with
      nr = radius - 2*Size
    and therefore
      x' = Trunc(CenterX + radius*sin(angle) - 2*Size*sin(angle) + 0.5)
    Because
      x = CenterX + radius*sin(angle)
    it follows that
      x' = Trunc(x - 2*Size*sin(angle) + 0.5)
    The term inside the Trunc() is always positive, so we're always
    rounding down. Therefore,
      x' = x - Floor(-2*Size*sin(angle) + 0.5)
    and thus
      x = x' + Floor(-2*Size*sin(angle) + 0.5)
    or, using Trunc,
      x = Trunc(x' - 2*Size*sin(angle) + 0.5)

    Now the problem is that we don't know then angle. Our ugly workaround,
    if an exact translation is requested, is to do brute force. The initial
    inexact hypothesis, computed using the above formula, is passed in as
    \c out. This iterates around the hypothesis until it finds a valid
    mapping. It may fail, though. This means that an inside point cannot
    be reached by moving to a particular outside point.

    \param config [in] Map config
    \param pt  [in] Point
    \param out [in] Initial hypothesis, [out] Result

    Derived from find.pas::CircularMoveOutside. */
static bool
findExactOutsideLocation(const GChartConfiguration& config, const GPoint pt, GPoint& out)
{
    // Maybe the hypothesis already is exact?
    if (config.getCanonicalLocation(out) == pt) {
        return true;
    }

    // Brute force
    for (int i = 1; i <= opt_circular_precision(); ++i) {
        int dx = 1;
        int dy = 1;
        enum { MoveDown, MoveLeft, MoveUp, MoveRight } where = MoveDown;
        while (1) {
            // Check this location
            GPoint t(out.x + dx, out.y + dy);
            if (config.getCanonicalLocation(t) == pt) {
                out = t;
                return true;
            }

            // Run around
            if (where == MoveDown) {
                if (dy + i == 0) { where = MoveLeft;  } else { --dy; }
            } else if (where == MoveLeft) {
                if (dx + i == 0) { where = MoveUp;    } else { --dx; }
            } else if (where == MoveUp) {
                if (dy == i)     { where = MoveRight; } else { ++dy; }
            } else {
                if (dx == i)     { break;             } else { ++dx; }
            }
        }
    }

    // Cannot find a mapping
    return false;
}

/********************************* GPoint ********************************/

/** Set this point by parsing a coordinate pair.
    \param str Coordinate
    \return true on success */
bool
GPoint::parseCoordinates(const string_t& str)
{
    int x, y;
    string_t::size_type p = str.find(',');
    if (p == str.npos)
        return false;

    if (!stringValue(str.substr(0, p), x) || !stringValue(str.substr(p+1), y))
        return false;

    this->x = x;
    this->y = y;
    return true;
}

/** Set this point by parsing a sector number.
    \param s Sector number as string
    \return true on success */
bool
GPoint::parseSectorNumber(const string_t& s)
{
    int sec;
    return stringValue(s, sec)
        && parseSectorNumber(sec);
}

/** Set this point by parsing a sector number.
    \param n Sector number as integer
    \return true on success */
bool
GPoint::parseSectorNumber(int n)
{
    /* Valid range is 100 .. 499 */
    if (n < 100 || n > 499)
        return false;

    /* Compute location */
    static const int secX[] = { -950,   50, -950,   50 };
    static const int secY[] = {  950,  950,  -50,  -50 };
    GPoint origin = chart_conf.getCenter();
    int newX = origin.x + secX[n / 100 - 1] + 100*((n % 100) / 10);
    int newY = origin.y + secY[n / 100 - 1] - 100*(n % 10);

    /* On map? */
    if (!GPoint(newX, newY).isOnMap())
        return false;

    /* Use it */
    this->x = newX;
    this->y = newY;

    return true;
}

/** Get sector number this point is in, as decimal number. This
    returns the same sector number as in PCC 1.x, which happens to be
    the same as the sectors on Trevor Fuson's VGAMAP for a
    standard-sized map.
    \return sector number (100..499), zero if point is not in any
    numbered sector */
int
GPoint::getSectorNumber() const
{
    /* outside map? */
    if (!isOnMap())
        return 0;

    int x = this->x - chart_conf.getCenter().x + 1000;
    int y = this->y - chart_conf.getCenter().y + 1000;

    /* outside known region? */
    if (x < 0 || y < 0 || x >= 2000 || y >= 2000)
        return 0;

    int major = 1;
    if (x >= 1000)
        major += 1;
    if (y < 1000)
        major += 2;

    return 100 * major
        + 10 * ((x % 1000) / 100)
        + (999 - (y % 1000)) / 100;
}

/************************** GChartConfiguration **************************/

GChartConfiguration::GChartConfiguration()
    : mode(Flat),
      center(2000, 2000),
      size(2000, 2000),
      from_hconfig(false)
{
    computeDerivedInformation();
}

/** Initialize from configuration. */
void
GChartConfiguration::initFromConfig()
{
    // Invariants (xref chartconfig.cc):
    // - center in range 500..4000
    // - size in range 500..4000
    // - size <= center

    // Load config file
    center.x = opt_map_center(1);
    center.y = opt_map_center(2);
    size.x   = opt_map_size(1);
    size.y   = opt_map_size(2);
    if (opt_map_kind() >= 0 && opt_map_kind() < 3) {
        // We assume a certain set of values. xref ValueEnumParser
        CTASSERT(Flat == 0);
        CTASSERT(Wrapped == 1);
        CTASSERT(Circular == 2);
        mode = Mode(opt_map_kind());
    }

    // Check host config
    if (host.isPHost()
        && config.AllowWraparoundMap.getSource() != ConfigOption::srcDefault
        && config.AllowWraparoundMap() != 0)
    {
        // It's a PHost game, and the AllowWraparoundMap option is set.
        // Copy the settings from pconfig.
        from_hconfig = true;
        mode         = Wrapped;
        center.x     = (config.WraparoundRectangle(3) + config.WraparoundRectangle(1)) / 2;
        center.y     = (config.WraparoundRectangle(4) + config.WraparoundRectangle(2)) / 2;
        size.x       = (config.WraparoundRectangle(3) - config.WraparoundRectangle(1));
        size.y       = (config.WraparoundRectangle(4) - config.WraparoundRectangle(2));
    } else {
        // Not a PHost game, or AllowWraparoundMap not set or disabled.
        // It could use external wrap, so don't change anything.
        from_hconfig = false;
    }
    computeDerivedInformation();
}

/** Modify configuration. */
void
GChartConfiguration::setConfig(Mode mode, GPoint center, GPoint size)
{
    this->mode         = mode;
    this->center       = center;
    this->size         = size;
    this->from_hconfig = false;
    computeDerivedInformation();
}

/** Check if point is on map. */
bool
GChartConfiguration::isOnMap(const GPoint& pt) const
{
    switch (mode) {
     case Flat:
        // Everything is on the map
        return true;

     case Wrapped:
        // Boundary behaviour consistent with PCC 1.x, PHost 3.3c, and EchoView
        return pt.x >= min.x && pt.x < max.x
            && pt.y >= min.y && pt.y < max.y;

     case Circular:
        // Check inside of bounding rectangle first, then distance.
        return pt.x >= min.x && pt.x <= max.x
            && pt.y >= min.y && pt.y <= max.y
            && distanceSquaredRaw(pt, center) <= int32_t(size.x)*size.x;
    }
    return false;
}

/** Check if point is a valid planet coordinate.
    Points may be on the map using isOnMap's strict rules,
    but by convention be treated as out-of-bounds,
    because they are mangled by the ExploreMap add-on before the
    host actually sees them. */
bool
GChartConfiguration::isValidPlanetCoordinate(const GPoint& pt) const
{
    return pt.x > 0 && pt.y > 0 && pt.x <= 9000 && pt.y <= 9000 && pt.isOnMap();
}

/** Given a point, get its canonical equivalent.
    If any kind of wrap is active, this performs the "wrap" step normally performed by the host. */
GPoint
GChartConfiguration::getCanonicalLocation(GPoint pt) const
{
    switch (mode) {
     case Flat:
        // No non-canonical locations
        break;

     case Wrapped:
        // Wrap into range
        if (pt.x < min.x) {
            pt.x += size.x;
        }
        if (pt.x >= max.x) {
            pt.x -= size.x;
        }
        if (pt.y < min.y) {
            pt.y += size.y;
        }
        if (pt.y >= max.y) {
            pt.y -= size.y;
        }
        break;

     case Circular:
        // Wrap into range
        int32_t dist = distanceSquaredRaw(pt, center);
        if (dist > int32_t(size.x)*size.x) {
            // pwrap formulas
            double radius = 2*size.x - std::sqrt(double(dist));
            double angle  = std::atan2(double(center.x - pt.x), double(center.y - pt.y));

            pt.x = int(radius * std::sin(angle) + center.x + 0.5);
            pt.y = int(radius * std::cos(angle) + center.y + 0.5);
        }
        break;
    }

    return pt;
}

/** Get canonical location of a point, simple version.
    This handles just rectangular wrap, where all instances of a location are equivalent. */
GPoint
GChartConfiguration::getSimpleCanonicalLocation(GPoint pt) const
{
    if (mode == Wrapped) {
        return getCanonicalLocation(pt);
    } else {
        return pt;
    }
}

/** Get nearest alias of a point, simple version.
    This handles just rectangular wrap, where all instances of a location are equivalent.
    Returns the instance of \c pt that is closest to \c a. */
GPoint
GChartConfiguration::getSimpleNearestAlias(GPoint pt, GPoint a) const
{
    if (mode == Wrapped) {
        if (2*(a.x - pt.x) > size.x)
            pt.x += size.x;
        if (2*(a.y - pt.y) > size.y)
            pt.y += size.y;
        if (2*(pt.x - a.x) > size.x)
            pt.x -= size.x;
        if (2*(pt.y - a.y) > size.y)
            pt.y -= size.y;
    }
    return pt;
}

/** Get number of map images that can map rectangles. */
int
GChartConfiguration::getNumRectangularImages() const
{
    switch (mode) {
     case Flat:
        return 1;
     case Circular:
        return 1;
     case Wrapped:
        return 9;
    }
    return 0;
}

/** Get number of map images that can map points. */
int
GChartConfiguration::getNumPointImages() const
{
    switch (mode) {
     case Flat:
        return 1;
     case Circular:
        return 2;
     case Wrapped:
        return 9;
    }
    return 0;
}

/** Compute outside location for a point inside the map.
    \param pt    [in] Point
    \param out   [out] Result will be produced here
    \param image [in] Index of map image to produce. 0=regular image.
    \param exact [in] True to request a perfect mapping, false to accept an inexact mapping
    \retval true this point could be mapped to the requested image
    \retval false this point could not be mapped */
bool
GChartConfiguration::getPointAlias(GPoint pt, GPoint& out, int image, bool exact) const
{
    switch (mode) {
     case Flat:
        // We have only one image
        if (image == 0) {
            out = pt;
            return true;
        } else {
            return false;
        }

     case Wrapped:
        // Point must be inside, and a supported image.
        if (!pt.isOnMap() || image < 0 || image >= 9) {
            return false;
        } else {
            out.x = pt.x + size.x * image_dx[image];
            out.y = pt.y + size.y * image_dy[image];
            return true;
        }

     case Circular:
        // Find radius.
        int32_t r;
        r = distanceSquaredRaw(center, pt);
        if (r > int32_t(size.x) * size.x) {
            // It's outside the permitted range
            return false;
        } else if (image == 0) {
            // Standard image
            out = pt;
            return true;
        } else if (image == 1) {
            // Outside image
            double rr = std::sqrt(double(r));
            if (rr > size.x - opt_circular_excess()) {
                // We might be able to map it
                double angle = std::atan2(double(pt.x - center.x), double(pt.y - center.y));
                out.x = int32_t(pt.x - 2*size.x*std::sin(angle) + 0.5);
                out.y = int32_t(pt.y - 2*size.x*std::cos(angle) + 0.5);
                if (exact) {
                    return findExactOutsideLocation(*this, pt, out);
                }
                return true;
            } else {
                // It's too far inside, so its mapping would be too far outside
                return false;
            }
        } else {
            // Error
            return false;
        }
    }
    return false;
}

/** Compute outside location for a point inside the map, simple version.
    This is well-suited to map known map objects in a fail-safe way.
    It does NOT map circular points to the outside. */
GPoint
GChartConfiguration::getSimplePointAlias(GPoint pt, int image) const
{
    switch (mode) {
     case Flat:
     case Circular:
        // No change for these
        break;

     case Wrapped:
        // Regular remapping
        pt.x += size.x * image_dx[image];
        pt.y += size.y * image_dy[image];
        break;
    }
    return pt;
}

void
GChartConfiguration::computeDerivedInformation()
{
    // Check ranges and force out-of-range values into range
    // xref chartconfig.cc
    if (center.x < 500 || center.x > 4000) {
        center.x = 2000;
        from_hconfig = false;
    }
    if (center.y < 500 || center.x > 4000) {
        center.y = 2000;
        from_hconfig = false;
    }
    if (size.x < 500 || size.x > center.x) {
        size.x = std::min(int(center.x), 2000);
        from_hconfig = false;
    }
    if (size.y < 500 || size.y > center.y) {
        size.y = std::min(int(center.y), 2000);
        from_hconfig = false;
    }

    // Compute derived information
    switch (mode) {
     case Flat:
     case Wrapped:
        min.x = center.x - size.x / 2;
        min.y = center.y - size.y / 2;
        max   = min + size;
        break;

     case Circular:
        size.y = size.x;
        min = center - size;
        max = center + size;

        // Update configuration. We limit wrap excess to 2/3 of the size,
        // which is 666 ly in the normal configuration.
        {
            int max_excess = size.x * 2/3;
            if (opt_circular_excess() > max_excess) {
                opt_circular_excess.set(max_excess);
                opt_circular_excess.setSource(ConfigOption::srcGame);
            }
        }

        // While we are at it, fix Precision as well. Range 0..20 is consistent
        // with PCC 1.x. Do not mark it for the game configuration, so it is
        // updated in whatever config file it was.
        if (opt_circular_precision() < 0) {
            opt_circular_precision.set(0);
        }
        if (opt_circular_precision() > 20) {
            opt_circular_precision.set(20);
        }
        break;
    }

    // Save to main config. loadUserPreferences() has set all options to srcUser.
    // This is wrong for map options which must always be srcGame. However, to
    // avoid creating game configuration files if users never even touched the
    // settings, we downgrade an option to srcDefault (which means it is not stored
    // in a config file) if all of the following holds:
    //   - it is srcUser (=user did not set it to srcGame)
    //   - it has the default value
    //   - it is being set to the default value
    if (opt_map_kind.getSource() <= ConfigOption::srcUser && opt_map_kind() == Flat && mode == Flat) {
        opt_map_kind.setSource(ConfigOption::srcDefault);
    } else {
        opt_map_kind.set(mode);
        opt_map_kind.markUpdated(ConfigOption::srcGame);
    }

    if (opt_map_center.getSource() <= ConfigOption::srcUser && opt_map_center(1) == 2000 && opt_map_center(2) == 2000 && center.x == 2000 && center.y == 2000) {
        opt_map_center.setSource(ConfigOption::srcDefault);
    } else {
        opt_map_center.set(1, center.x);
        opt_map_center.set(2, center.y);
        opt_map_center.markUpdated(ConfigOption::srcGame);
    }

    if (opt_map_size.getSource() <= ConfigOption::srcUser && opt_map_size(1) == 2000 && opt_map_size(2) == 2000 && size.x == 2000 && size.y == 2000) {
        opt_map_size.setSource(ConfigOption::srcDefault);
    } else {
        opt_map_size.set(1, size.x);
        opt_map_size.set(2, size.y);
        opt_map_size.markUpdated(ConfigOption::srcGame);
    }
}

/******************************* Utilities *******************************/

/** Get minimum distance between two points, squared. */
long
distanceSquared(GPoint a, GPoint b)
{
    GPoint c = b.getSimpleNearestAlias(a);
    return distanceSquaredRaw(a, c);
}

/** Get distance between two points, squared. */
long
distanceSquaredRaw(GPoint a, GPoint b)
{
    long x = a.x - b.x;
    long y = a.y - b.y;
    return x*x + y*y;
}

/** Initialize. Assumes configuration has been read. */
static void
initCoordinates()
{
    chart_conf.initFromConfig();
}

/** Output GPoint to a stream. */
std::ostream&
operator<<(std::ostream& os, GPoint pt)
{
    return os << '(' << pt.x << ", " << pt.y << ')';
}

// Constraints: after mod_config, after mod_pref
const ModuleInitializer mod_coord[] = {
    // Initialize this module whenever configuration has been read.
    // Configuration is modified in three cases:
    { OnStartup | OnEnterNoGame | OnEnterGame, initCoordinates },
    { 0, 0 }
};
