/**
  *  \file game/save-msg.cc
  *  \brief Functions for Saving Message Files
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/save.h"
#include "game/msg.h"
#include "io/dirs.h"
#include "io/io.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/struct.h"
#include "util/misc.h"

/** Save message outbox, file format 3.5.
    \param outbox  Message box to save
    \param dir     Target directory
    \param player  Player owning the file (should be same as
                   outbox.getSender()) */
void
saveOutbox35(GOutbox& outbox, Directory& dir, int player)
{
    Ptr<Stream> f = dir.openFile(format("mess35%d.dat", player), Stream::C_CREATE);
    char header_buffer[19];
    zeroFill(header_buffer);
    storeUint16(header_buffer, outbox.getCount());
    if (!outbox.getCount()) {
        f->writeT(header_buffer, 2);
    } else {
        f->writeT(header_buffer, 19);
        for (int i = 0; i < outbox.getCount(); ++i) {
            /* FIXME: Winplan has a limit of 600 characters. PCC 1.x allows
               to exceed that limit if Winplan is not in use. Right now,
               we enforce the limit. */
            string_t text = outbox.getRawText(i);
            GPlayerSet recv = outbox.getReceiverMask(i);
            encodeMessage(text);
            for (string_t::size_type n = text.length(); n > 0; --n)
                if (text[n-1] == 26)
                    text.insert(n, 1, 23);

            char message_buffer[616];
            message_buffer[0] = 0;     // filler
            message_buffer[1] = '1';   // valid flag
            for (int p = 1; p <= 12; ++p)
                message_buffer[1+p] = (recv.contains(p)) ? '1' : '0';
            storeUint16(message_buffer + 14, 600);

            std::fill_n(message_buffer + 16, 600, char(32+13));
            copy_n(text.begin(), std::min(std::size_t(text.length()), std::size_t(600)), message_buffer + 16);
            f->writeT(message_buffer, 616);
        }
    }
}

/** Save message outbox, file format 3.0.
    \param outbox  Message box to save
    \param dir     Target directory
    \param player  Player owning the file (should be same as
                   outbox.getSender()) */
void
saveOutbox(GOutbox& outbox, Directory& dir, int player)
{
    /* count messages */
    int nmessages = 0;
    for (int i = 0; i < outbox.getCount(); ++i) {
        GPlayerSet recv = outbox.getReceiverMask(i);
        for (int p = 1; p <= NUM_PLAYERS+1; ++p)
            if (recv.contains(p))
                ++nmessages;
    }

    // FIXME: some kind of overflow check here?

    /* Allocate buffer for message headers. We leave room for at least
       50 messages after those we wrote, but at least 150 in total. 
       This is to help programs that don't rewrite the message file
       from scratch. PLANETS.EXE has a fixed limit of 50 here; VPA
       seems to relocate messages dynamically when the directory grows
       too large. */
    int nslots = nmessages + 50;
    if (nslots < 150)
        nslots = 150;

    int msgh_size = nslots * TOutgoingMessageHeader::size + 2;
    Array<char> buffer(msgh_size);
    buffer[0U] = buffer[1U] = 0;

    /* write file */
    Ptr<Stream> f = dir.openFile(format("mess%d.dat", player), Stream::C_CREATE);
    f->writeT(buffer.ptr(), msgh_size);

    uint32 output_index = msgh_size + 1;  // +1 for BASIC
    int    header_index = 2;
    for (int i = 0; i < outbox.getCount(); ++i) {
        GPlayerSet recv = outbox.getReceiverMask(i);
        for (int p = 1; p <= NUM_PLAYERS+1; ++p) {
            if (recv.contains(p)) {
                /* send one message */
                string_t text_to_send = outbox.getSendPrefix(i, p) + outbox.getRawText(i);
                encodeMessage(text_to_send);

                TOutgoingMessageHeader h;
                h.address = output_index;
                h.length  = text_to_send.length();
                h.from    = outbox.getSender();
                h.to      = p;
                storeStructure(buffer.ptr() + header_index, h);
                f->writeT(text_to_send.data(), text_to_send.size());
                header_index += TOutgoingMessageHeader::size;
                output_index += text_to_send.length();
            }
        }
    }

    ASSERT(header_index <= msgh_size);
    storeUint16(buffer.ptr(), nmessages);
    f->seek(0);
    f->writeT(buffer, msgh_size);
}

