/**
  *  \file int/if/shippred.cc
  *  \brief Interpreter Game Interface: Ship Prediction
  *
  *  \todo This is bad abstraction: we have GShip, GShipTurnPredictor,
  *  and this one, all with different accessor interfaces. It would
  *  make a lot of sense to merge them all.
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/shippred.h"
#include "game/global.h"
#include "int/if/ifutil.h"
#include "int/args.h"
#include "game/ship.h"
#include "game/mission.h"

/** Constructor. Create predictor. */
IntShipPredictor::IntShipPredictor(int id)
    : pred(getCurrentTurn().getCurrentUniverse(), id),
      num_xys(0),
      fuel_xys(0),
      turn_fuel(0),
      have_fuel(true)
{ }

IntShipPredictor::~IntShipPredictor()
{ }

/** Handle "SetWaypoint"-style command.
    \param args Args passed to command. */
void
IntShipPredictor::setWaypoint(IntArgBlock& args)
{
    // Read args
    int32_t x = 2000, y = 2000;
    args.checkArgumentCount(2);
    checkIntArg(x, args.getNext(), 0, 10000);
    checkIntArg(y, args.getNext(), 0, 10000);

    // Set waypoint
    pred.setWaypoint(GPoint(x, y));
}

/** Handle "WaitOneTurn"-style command. Advances the time and updates
    all values. */
void
IntShipPredictor::predictTurn()
{
    pred.computeTurn();

    // Remember ship position
    if (num_xys < MAX_XYS) {
        GPoint pt = pred.getPosition();
        if (num_xys == 0 || pt != positions[num_xys-1]) {
            positions[num_xys++] = pt;
        }
    }

    // Remember that we had fuel
    if (have_fuel && pred.getFuel() >= 0) {
        turn_fuel = pred.getNumTurns();
        fuel_xys = num_xys;
    }

    // Avoid running out of fuel
    if (pred.getFuel() < 0) {
        pred.setFuel(0);
        have_fuel = false;
    }
}

bool
IntShipPredictor::predictInstruction(const string_t& name, IntArgBlock& args)
{
    if (name == "MOVETO") {
        setWaypoint(args);
        int n = 0;
        while (n < GShipTurnPredictor::MOVEMENT_TIME_LIMIT && !pred.isAtWaypoint()) {
            predictTurn();
            ++n;
        }
        return true;
    } else if (name == "SETWAYPOINT") {
        setWaypoint(args);
        return true;
    } else if (name == "MOVETOWARDS") {
        setWaypoint(args);
        predictTurn();
        return true;
    } else if (name == "WAITONETURN") {
        predictTurn();
        return true;
    } else if (name == "SETSPEED") {
        args.checkArgumentCount(1);
        int32_t speed = 0;
        checkIntArg(speed, args.getNext(), 0, 9);
        pred.setSpeed(speed);
        return true;
    } else if (name == "SETFCODE") {
        args.checkArgumentCount(1);
        string_t fc;
        checkStringArg(fc, args.getNext());
        pred.setFCode(fc);
        return true;
    } else if (name == "SETMISSION") {
        args.checkArgumentCount(1, 3);
        int32_t m = 0, i = 0, t = 0;
        checkIntArg(m, args.getNext(), 0, 10000);
        checkIntArg(i, args.getNext(), 0, 10000);
        checkIntArg(t, args.getNext(), 0, 10000);
        pred.setMission(m, i, t);

        const GMission* msn = GMissionList::getInstance().getMissionByNumber(m, GPlayerSet(pred.getRealOwner()));
        if (msn != 0 && msn->hasFlag(GMission::mf_Waypoint) && pred.getUniverse().ty_any_ships.isValidIndex(i)) {
            const GShip& sh = pred.getUniverse().getShip(i);

            // FIXME: handle THost where intercept does not cross the seam?
            pred.setWaypoint(sh.getPos().getSimpleNearestAlias(pred.getPosition()));
        }
        return true;
    } else {
        return true;
    }
}
