/**
  *  \file int/intglobal.cc
  *  \brief Interpreter: Global Definitions
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/intglobal.h"
#include "int/names.h"
#include "int/data.h"
#include "game/struct.h"
#include "cpluslib/assert.h"
#include "cpluslib/misc.h"
#include "int/context.h"
#include "util/modinit.h"
#include "int/compound.h"
#include "int/file.h"
#include "int/mutex.h"

/*************************** Exported variables **************************/

/** Names of all planet properties. */
IntVariableNames planet_property_names;

/** Names of all ship properties. */
IntVariableNames ship_property_names;

/** Names of all global variables. */
IntVariableNames global_property_names;

/** All global values. */
IntDataSegment global_values;


/*************************** Accessor Functions **************************/

/** Planet property values. */
static IntDataSegment planet_properties[NUM_PLANETS];

/** Ship property values. */
static IntDataSegment ship_properties[NUM_SHIPS];


/** Initialize interpreter globals. Call this once per program invocation. */
static void
initInterpreter()
{
    /* @q Comment:Str (Ship Property, Planet Property)
       User comment.
       This is the comment that can be edited with <kbd>F9</kbd>.
       @assignable */
    ship_property_names.add("COMMENT");
    planet_property_names.add("COMMENT");
    for (char c = 'A'; c < 'Z'; ++c)
        global_property_names.add(string_t(1, c));

    /* @q System.Err:Str (Global Variable)
       Error message.
       If a command within a {Try} statement generates an error,
       the error message is stored in this variable.
       The %Else part of the %Try statement can therefore look at the message,
       or re-throw the error using {Abort}.

       If a local variable %System.Err is visible, the error message is stored in that
       instead of the global one.

       @diff In PCC 1.x, %System.Err is a global property; the error message is always
       stored in the global property, and a local %System.Err is ignored.
       @assignable */
    global_property_names.add("SYSTEM.ERR");
}

/** Discard interpreter's game-specific data. Call this after closing
    down the game. */
static void
clearInterpreterGameData()
{
    // Clear all properties
    for (unsigned i = 0; i < countof(planet_properties); ++i)
        planet_properties[i].clear();
    for (unsigned i = 0; i < countof(ship_properties); ++i)
        ship_properties[i].clear();
}

/** Get property data segment for a ship. */
IntDataSegment&
getShipProperties(int sid)
{
    ASSERT(sid > 0 && sid <= NUM_SHIPS);
    return ship_properties[sid-1];
}

/** Get property data segment for a planet. */
IntDataSegment&
getPlanetProperties(int pid)
{
    ASSERT(pid > 0 && pid <= NUM_PLANETS);
    return planet_properties[pid-1];
}

/** Define a global procedure.
    \param name Name of the procedure
    \param proc C++ procedure to execute */
void
defineGlobalProcedure(const char* name, void proc(IntExecutionContext& exc, IntArgBlock& args))
{
    IntVariableNames::index_t i = global_property_names.addMaybe(name);
    global_values.setNew(i, new IntSimpleProcedureValue(proc));
}

/** Define a global value.
    \param name Name of the value
    \param proc value */
void
defineGlobalValue(const char* name, IntValue* value)
{
    IntVariableNames::index_t i;
    try {
        i = global_property_names.addMaybe(name);
    }
    catch (...) {
        delete value;
        throw;
    }
    global_values.setNew(i, value);
}

const ModuleInitializer mod_int[] = {
    // On startup, initialize.
    { OnStartup, initInterpreter },
    { OnStartup, initInterpreterFileInterface },
    { OnStartup, initInterpreterMutexInterface },
    // After closing a game, discard game data. However, this does not discard additional
    // properties created by the game.
    { OnLeftGame, clearInterpreterGameData },
    { 0, 0 },
};
