/**
  *  \file io/zip.cc
  *  \brief Zip file access.
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "io/zip.h"
#include "io/io.h"
#include "io/zipstream.h"
#include "util/except.h"
#include "util/nls.h"
#include "util/recode.h"

namespace {
    // General flags
    enum {
        gfEncrypted       = 1,
        gfMethodExtra1    = 2,
        gfMethodExtra2    = 4,
        gfUnicodeName     = 2048,
        gfUpdated         = 32768,         // not documented by PKWARE, but in use
        gfKnown           = gfEncrypted + gfMethodExtra1 + gfMethodExtra2 + gfUnicodeName + gfUpdated
    };

    // Compression methods
    enum {
        cmStored   = 0,
        cmDeflated = 8
    };
}

/** Index entry. Stores information gathered from the file member header. */
struct ZipFile::IndexEntry {
    string_t name;
    uint16_t method;
    Stream::fpos_t start;
    Stream::fpos_t compressed_size;
    Stream::fpos_t uncompressed_size;

    IndexEntry(const string_t& name, uint16_t method,
               Stream::fpos_t start,
               Stream::fpos_t compressed_size,
               Stream::fpos_t uncompressed_size)
        : name(name),
          method(method),
          start(start),
          compressed_size(compressed_size),
          uncompressed_size(uncompressed_size)
        { }
};

/** Zip "DirectoryEntry" implementation. */
class ZipFile::ZipDirEntry : public DirectoryEntry {
    IndexEntry& entry;
    Ptr<ZipFile> parent;
 public:
    ZipDirEntry(IndexEntry& entry, Ptr<ZipFile> parent);
    ~ZipDirEntry();
    void getInfo();
    Ptr<Directory> openDir();
    Ptr<Stream> openFile(int mode);
    Ptr<Directory> openParent();
    void erase();
    void renameTo(string_t name);
    void createAsDirectory();
};

/** Zip "Enumeration<DirectoryEntry>" implementation. */
class ZipFile::ZipDirEnum : public Enumeration<Ptr<DirectoryEntry> > {
    Ptr<ZipFile> parent;
    std::size_t current;
 public:
    ZipDirEnum(Ptr<ZipFile> parent);
    ~ZipDirEnum();
    bool hasMoreElements();
    Ptr<DirectoryEntry> getNextElement();
};

/** Zip "Stream" implementation for a stored member. This cannot be
    just a simple SubStream (although it conceptually isn't much more)
    because it has to manage the lifetime of the file. */
class ZipFile::ZipStoredMember : public Stream {
    Ptr<Stream> file;      /**< Master file (for lifetime control). */
    ChildStream cs;        /**< Isolation against other users. */
    SubStream ss;          /**< Limit checking. */
    string_t name;
 public:
    ZipStoredMember(Ptr<Stream> file, IndexEntry& entry);
    ~ZipStoredMember();

    std::size_t read(char* buffer, std::size_t max);
    std::size_t write(const char* buffer, std::size_t max);
    int getCapabilities() const;
    void seek(fpos_t pos);
    fpos_t getPos();
    fpos_t getSize();
    string_t getName();
};

/** Zip "Stream" implementation for a deflated member. */
class ZipFile::ZipDeflatedMember : public Stream {
    Ptr<Stream> file;      /**< Master file (for lifetime control). */
    ChildStream cs;        /**< Isolation against other users. */
    SubStream ss;          /**< Limit checking. */
    ZipStream zs;          /**< Decompression. */
    string_t name;
 public:
    ZipDeflatedMember(Ptr<Stream> file, IndexEntry& entry);
    ~ZipDeflatedMember();

    std::size_t read(char* buffer, std::size_t max);
    std::size_t write(const char* buffer, std::size_t max);
    int getCapabilities() const;
    void seek(fpos_t pos);
    fpos_t getPos();
    fpos_t getSize();
    string_t getName();
};


/************************** ZipFile::ZipDirEntry *************************/

/** Constructor.
    \param entry Index entry for this directory entry
    \param parent Link to containing zip file. */
ZipFile::ZipDirEntry::ZipDirEntry(IndexEntry& entry, Ptr<ZipFile> parent)
    : DirectoryEntry(entry.name),
      entry(entry),
      parent(parent)
{ }

ZipFile::ZipDirEntry::~ZipDirEntry()
{ }

void
ZipFile::ZipDirEntry::getInfo()
{
    setInfo(entry.uncompressed_size, FT_FILE);
}

Ptr<Directory>
ZipFile::ZipDirEntry::openDir()
{
    throw FileOpenException(getName(), _("No such directory"));
}

Ptr<Stream>
ZipFile::ZipDirEntry::openFile(int mode)
{
    if (mode & (Stream::C_CREATE | Stream::C_WRITE)) {
        throw FileOpenException(getName(), _("Cannot modify archive file"));
    } else if (entry.method == cmStored) {
        return new ZipStoredMember(parent->file, entry);
    } else if (entry.method == cmDeflated) {
        return new ZipDeflatedMember(parent->file, entry);
    } else {
        throw FileOpenException(getName(), _("Invalid compression mode"));
    }
}

Ptr<Directory>
ZipFile::ZipDirEntry::openParent()
{
    return parent;
}

void
ZipFile::ZipDirEntry::erase()
{
    throw FileOpenException(getName(), _("Cannot modify archive file"));
}

void
ZipFile::ZipDirEntry::renameTo(string_t /*name*/)
{
    throw FileOpenException(getName(), _("Cannot modify archive file"));
}

void
ZipFile::ZipDirEntry::createAsDirectory()
{
    throw FileOpenException(getName(), _("Cannot modify archive file"));
}

/************************** ZipFile::ZipDirEnum **************************/

/** Constructor.
    \param parent Link to containing zip file. */
ZipFile::ZipDirEnum::ZipDirEnum(Ptr<ZipFile> parent)
    : parent(parent),
      current(0)
{ }

ZipFile::ZipDirEnum::~ZipDirEnum()
{ }

bool
ZipFile::ZipDirEnum::hasMoreElements()
{
    while (parent->index.size() <= current && parent->readNextEntry()) {
        /* nix */
    }
    return parent->index.size() > current;
}

Ptr<DirectoryEntry>
ZipFile::ZipDirEnum::getNextElement()
{
    return new ZipDirEntry(*parent->index[current++], parent);
}

/************************ ZipFile::ZipStoredMember ***********************/

/** Constructor.
    \param parent Link to containing zip file.
    \param entry Index entry for this member */
ZipFile::ZipStoredMember::ZipStoredMember(Ptr<Stream> file, IndexEntry& entry)
    : file(file),
      cs(*file),
      ss(cs),
      name(entry.name)
{
    ss.setRange(entry.start, entry.compressed_size);
    ss.seek(0);
}

ZipFile::ZipStoredMember::~ZipStoredMember()
{ }

std::size_t
ZipFile::ZipStoredMember::read(char* buffer, std::size_t max)
{
    return ss.read(buffer, max);
}

std::size_t
ZipFile::ZipStoredMember::write(const char* /*buffer*/, std::size_t /*max*/)
{
    return notWritable();
}

int
ZipFile::ZipStoredMember::getCapabilities() const
{
    return C_READ | C_SEEK;
}

void
ZipFile::ZipStoredMember::seek(fpos_t pos)
{
    ss.seek(pos);
}

Stream::fpos_t
ZipFile::ZipStoredMember::getPos()
{
    return ss.getPos();
}

Stream::fpos_t
ZipFile::ZipStoredMember::getSize()
{
    return ss.getSize();
}

string_t
ZipFile::ZipStoredMember::getName()
{
    return Directory::makeFileName(file->getName(), name);
}

/*********************** ZipFile::ZipDeflatedMember **********************/

/** Constructor.
    \param parent Link to containing zip file.
    \param entry Index entry for this member */
ZipFile::ZipDeflatedMember::ZipDeflatedMember(Ptr<Stream> file, IndexEntry& entry)
    : file(file),
      cs(*file),
      ss(cs),
      zs(ss, false),
      name(entry.name)
{
    ss.setRange(entry.start, entry.compressed_size);
    ss.seek(0);
}

ZipFile::ZipDeflatedMember::~ZipDeflatedMember()
{ }

std::size_t
ZipFile::ZipDeflatedMember::read(char* buffer, std::size_t max)
{
    return zs.read(buffer, max);
}

std::size_t
ZipFile::ZipDeflatedMember::write(const char* /*buffer*/, std::size_t /*max*/)
{
    return notWritable();
}

int
ZipFile::ZipDeflatedMember::getCapabilities() const
{
    return zs.getCapabilities();
}

void
ZipFile::ZipDeflatedMember::seek(fpos_t pos)
{
    zs.seek(pos);
}

Stream::fpos_t
ZipFile::ZipDeflatedMember::getPos()
{
    return zs.getPos();
}

Stream::fpos_t
ZipFile::ZipDeflatedMember::getSize()
{
    return zs.getSize();
}

string_t
ZipFile::ZipDeflatedMember::getName()
{
    return Directory::makeFileName(file->getName(), name);
}

/******************************** ZipFile ********************************/

/** Constructor. Use create(Ptr<Stream>) to create instances of this class. */
ZipFile::ZipFile(Ptr<Stream> file, int options)
    : file(file),
      view(*file),
      options(options),
      indexer_reached_end(false),
      had_unsupported_files(false),
      had_unsupported_feature(false)
{ }

ZipFile::~ZipFile()
{ }

/** Read next entry from Zip file.
    \retval true I have read something, successfully. It may or may not have been added to the index.
    \retval false I have stopped parsing, at end of file or at an item I don't understand */
bool
ZipFile::readNextEntry()
{
    /* Read block type (unless end already reached) */
    char buffer[100];
    if (indexer_reached_end || had_unsupported_feature || view.read(buffer, 4) != 4) {
        indexer_reached_end = true;
        return false;
    }
    uint32_t block = getUint32(buffer);

    if (block == 0x04034B50) {
        /* Local Header */
        view.readT(buffer, 26);
        // uint16_t min_version = getUint16(buffer+0);
        uint16_t flags       = getUint16(buffer+2);
        uint16_t method      = getUint16(buffer+4);
        // uint32_t mtime       = getUint32(buffer+6);
        // uint32_t crc         = getUint32(buffer+10);
        uint32_t compsize    = getUint32(buffer+14);
        uint32_t uncompsize  = getUint32(buffer+18);
        uint32_t namelength  = getUint16(buffer+22);
        uint32_t extralength = getUint16(buffer+24);

        /* If this header is not supported, stop parsing */
        if ((flags & ~gfKnown) != 0) {
            had_unsupported_feature = true;
            return false;
        }

        /* Read file name */
        string_t name;
        getString(view, name, namelength);
        string_t::size_type n = name.rfind('/');
        if (n != name.npos && ((options & KeepPaths) == 0 || n+1 == name.size())) {
            /* Clear name if this is a directory, remove path part if configured so. */
            name.erase(0, n+1);
        }
        if ((flags & gfUnicodeName) == 0) {
            name = convertGameToUtf8(name);
        }

        /* Skip extra field */
        if (extralength != 0) {
            view.seek(view.getPos() + extralength);
        }

        /* If it is supported, build index entry */
        if (name.size() == 0) {
            /* Directory */
        } else if ((flags & gfEncrypted) == 0 && (method == cmStored || method == cmDeflated)) {
            /* Supported file */
            if (!findEntryByName(name, 0, index.size())) {
                index.push_back(new IndexEntry(name, method, view.getPos(), compsize, uncompsize));
            }
        } else {
            /* Unsupported */
            had_unsupported_files = true;
        }
        view.seek(view.getPos() + compsize);
        return true;
    } else if (block == 0x02014B50 || block == 0x06054B50) {
        /* Central Directory, or End of Central Directory */
        indexer_reached_end = true;
        return false;
    } else {
        had_unsupported_feature = true;
        return false;
    }
}

ZipFile::IndexEntry*
ZipFile::findEntryByName(const string_t& name, std::size_t start, std::size_t end)
{
    for (std::size_t i = start; i < end; ++i) {
        if (index[i]->name == name) {
            return index[i];
        }
    }
    return 0;
}

Ptr<DirectoryEntry>
ZipFile::getEntryByName(string_t name)
{
    // Look in existing index
    std::size_t size = index.size();
    IndexEntry* xe = findEntryByName(name, 0, size);

    // Not found, so read more
    while (xe == 0 && readNextEntry()) {
        std::size_t nsize = index.size();
        xe = findEntryByName(name, size, nsize);
        size = nsize;
    }

    // Found?
    if (xe) {
        return new ZipDirEntry(*xe, this);
    } else {
        return 0;
    }
}

Ptr<Enumeration<Ptr<DirectoryEntry> > >
ZipFile::getEntries()
{
    return new ZipDirEnum(this);
}

Ptr<Directory>
ZipFile::getParent()
{
    return 0;
}

string_t
ZipFile::getName()
{
    return file->getName();
}

/** Constructor. Open a file as Zip file for reading.
    \param file A Zip file */
Ptr<ZipFile>
ZipFile::open(Ptr<Stream> file, int options)
{
    return new ZipFile(file, options);
}
